import { badgeServices, redisTtl } from "@config/environment";
import { fetch, redis } from "bun";

const DISCORD_BADGES = {
	// User badges
	HYPESQUAD: 2 << 2,
	HYPESQUAD_ONLINE_HOUSE_1: 2 << 6,
	HYPESQUAD_ONLINE_HOUSE_2: 2 << 7,
	HYPESQUAD_ONLINE_HOUSE_3: 2 << 8,

	STAFF: 2 << 0,
	PARTNER: 2 << 1,
	CERTIFIED_MODERATOR: 2 << 18,

	VERIFIED_DEVELOPER: 2 << 17,
	ACTIVE_DEVELOPER: 2 << 22,

	PREMIUM_EARLY_SUPPORTER: 2 << 9,

	BUG_HUNTER_LEVEL_1: 2 << 3,
	BUG_HUNTER_LEVEL_2: 2 << 14,

	// Bot badges
	SUPPORTS_COMMANDS: 2 << 23,
	USES_AUTOMOD: 2 << 24,
};

const DISCORD_BADGE_DETAILS = {
	HYPESQUAD: {
		tooltip: "HypeSquad Events",
		icon: "/public/badges/discord/HYPESQUAD.svg",
	},
	HYPESQUAD_ONLINE_HOUSE_1: {
		tooltip: "HypeSquad Bravery",
		icon: "/public/badges/discord/HYPESQUAD_ONLINE_HOUSE_1.svg",
	},
	HYPESQUAD_ONLINE_HOUSE_2: {
		tooltip: "HypeSquad Brilliance",
		icon: "/public/badges/discord/HYPESQUAD_ONLINE_HOUSE_2.svg",
	},
	HYPESQUAD_ONLINE_HOUSE_3: {
		tooltip: "HypeSquad Balance",
		icon: "/public/badges/discord/HYPESQUAD_ONLINE_HOUSE_3.svg",
	},

	STAFF: {
		tooltip: "Discord Staff",
		icon: "/public/badges/discord/STAFF.svg",
	},
	PARTNER: {
		tooltip: "Discord Partner",
		icon: "/public/badges/discord/PARTNER.svg",
	},
	CERTIFIED_MODERATOR: {
		tooltip: "Certified Moderator",
		icon: "/public/badges/discord/CERTIFIED_MODERATOR.svg",
	},

	VERIFIED_DEVELOPER: {
		tooltip: "Verified Bot Developer",
		icon: "/public/badges/discord/VERIFIED_DEVELOPER.svg",
	},
	ACTIVE_DEVELOPER: {
		tooltip: "Active Developer",
		icon: "/public/badges/discord/ACTIVE_DEVELOPER.svg",
	},

	PREMIUM_EARLY_SUPPORTER: {
		tooltip: "Premium Early Supporter",
		icon: "/public/badges/discord/PREMIUM_EARLY_SUPPORTER.svg",
	},

	BUG_HUNTER_LEVEL_1: {
		tooltip: "Bug Hunter (Level 1)",
		icon: "/public/badges/discord/BUG_HUNTER_LEVEL_1.svg",
	},
	BUG_HUNTER_LEVEL_2: {
		tooltip: "Bug Hunter (Level 2)",
		icon: "/public/badges/discord/BUG_HUNTER_LEVEL_2.svg",
	},

	SUPPORTS_COMMANDS: {
		tooltip: "Supports Commands",
		icon: "/public/badges/discord/SUPPORTS_COMMANDS.svg",
	},
	USES_AUTOMOD: {
		tooltip: "Uses AutoMod",
		icon: "/public/badges/discord/USES_AUTOMOD.svg",
	},
}

export async function fetchBadges(
	userId: string,
	services: string[],
	options?: FetchBadgesOptions,
): Promise<BadgeResult> {
	const { nocache = false, separated = false } = options ?? {};
	const results: Record<string, Badge[]> = {};

	await Promise.all(
		services.map(async (service) => {
			const entry = badgeServices.find(
				(s) => s.service.toLowerCase() === service.toLowerCase(),
			);
			if (!entry) return;

			const serviceKey = service.toLowerCase();
			const cacheKey = `badges:${serviceKey}:${userId}`;

			if (!nocache) {
				const cached = await redis.get(cacheKey);
				if (cached) {
					try {
						const parsed: Badge[] = JSON.parse(cached);
						results[serviceKey] = parsed;
						return;
					} catch {
						// corrupted cache, proceed with fetch :p
					}
				}
			}

			let url: string;
			let headers: Record<string, string> = {};
			if (typeof entry.url === "function") {
				url = entry.url(userId);
			} else {
				url = entry.url;
			}

			if (entry.authType === "DISCORD") {
				headers = {
					Authorization: `Bot ${process.env.DISCORD_TOKEN}`
				};
			}

			try {
				const res = await fetch(url, { headers });
				if (!res.ok) return;
				const data = await res.json();

				const result: Badge[] = [];

				switch (serviceKey) {
					case "discord": {
						if (data.avatar.startsWith("a_")) {
							result.push({
								tooltip: "Discord Nitro",
								badge: "/public/badges/discord/NITRO.svg",
							});
						}

						for (const [flag, bitwise] of Object.entries(DISCORD_BADGES)) {
							if (data.flags & bitwise) {
								const badge = DISCORD_BADGE_DETAILS[flag as keyof typeof DISCORD_BADGE_DETAILS];
								result.push({
									tooltip: badge.tooltip,
									badge: badge.icon,
								});
							}
						}
						break;
					}

					case "vencord":
					case "equicord": {
						const userBadges = data[userId];
						if (Array.isArray(userBadges)) {
							for (const b of userBadges) {
								result.push({
									tooltip: b.tooltip,
									badge: b.badge,
								});
							}
						}
						break;
					}

					case "nekocord": {
						const userBadgeIds = data.users?.[userId]?.badges;
						if (Array.isArray(userBadgeIds)) {
							for (const id of userBadgeIds) {
								const badgeInfo = data.badges?.[id];
								if (badgeInfo) {
									result.push({
										tooltip: badgeInfo.name,
										badge: badgeInfo.image,
									});
								}
							}
						}
						break;
					}

					case "reviewdb": {
						for (const b of data) {
							if (b.discordID === userId) {
								result.push({
									tooltip: b.name,
									badge: b.icon,
								});
							}
						}
						break;
					}
				}

				if (result.length > 0) {
					results[serviceKey] = result;
					if (!nocache) {
						await redis.set(cacheKey, JSON.stringify(result));
						await redis.expire(cacheKey, redisTtl);
					}
				}
			} catch (_) { }
		}),
	);

	if (separated) return results;

	const combined: Badge[] = [];
	for (const group of Object.values(results)) {
		combined.push(...group);
	}
	return combined;
}
